/*
 * A memory pool allocating a region at a time from the system.
 *
 * Copyright (C) 2021, 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include "memory_incremental.h"

#include <ipc/mem_ipc.h>

#include <stdlib.h>



/* Initialise the memory pool with an optional 'limit' in pages. */

MemoryIncremental::MemoryIncremental(unsigned int limit, offset_t region_size)
: _limit(limit), _region_size(region_size)
{
    _limited = true;
}

MemoryIncremental::MemoryIncremental()
: _region_size(PAGE_SIZE)
{
    _limited = false;
}

/* Allocate a block of the given 'size'. */

Region *MemoryIncremental::allocate(offset_t size)
{
    /* Attempt to allocate aligned memory. */

    void *current = NULL;
    l4re_ds_t ds;

    /* Make the size appropriate for the invocation. */

    size = round_multiple(size, PAGE_SIZE);

    /* Use allocation permitting executable mapping of the memory.
       NOTE: A collection of blocks needs to be allocated, and a collection of
             dataspaces might also be employed. */

    if (ipc_allocate_align(size, L4RE_RM_F_SEARCH_ADDR | L4RE_RM_F_RWX, page_order(size), &current, &ds))
        return NULL;

    return new Region((offset_t) current, (offset_t) current + size);
}

/* Deallocate the given region. */

void MemoryIncremental::deallocate(Region *region)
{
    ipc_detach_dataspace((void *) region->start);
    delete region;
}

/* Allocate a new region of the given 'size' rounded to the nearest page. */

Region *MemoryIncremental::region(offset_t size)
{
    std::lock_guard<std::mutex> guard(_lock);

    offset_t rounded = round(size, PAGE_SIZE);
    offset_t pages = rounded / PAGE_SIZE;

    /* Check for sufficient pages. */

    if (!_limited || (_limit >= pages))
    {
        /* Attempt to allocate aligned memory. */

        Region *region = allocate(rounded);

        if (region == NULL)
            return NULL;

        if (_limited)
            _limit -= pages;

        return region;
    }

    /* Return no region without sufficient pages. */

    else
        return NULL;
}

/* Allocate a new region having the default region size rounded to the nearest
   page. */

Region *MemoryIncremental::region()
{
    return region(_region_size);
}

/* Release the allocated 'region'. */

void MemoryIncremental::release(Region *region)
{
    std::lock_guard<std::mutex> guard(_lock);

    offset_t rounded = round(region->size(), PAGE_SIZE);
    offset_t pages = rounded / PAGE_SIZE;

    if (_limited)
        _limit += pages;

    deallocate(region);
}

// vim: tabstop=4 expandtab shiftwidth=4
